import enum
import uuid
from dataclasses import dataclass

__all__ = ["Tag", "EnvType", "AnnotationType", "GoldenSetInteraction", "Step", "StepType"]

from datetime import datetime
from typing import Dict, Any, Union


class Tag(str, enum.Enum):
    """
    Namespace for useful tags that deepchecks case use
    You can use `dc_client.set_tag()` to pass user tags to deepchecks

    USER_INPUT
        Relevant only for auto_collect=True, should contain the input as the user supply it

    INFORMATION_RETRIEVAL
        Relevant only for auto_collect=True, should contain the "information retrieval" if exist

    USER_ID
        The external user that used the AI model

    EXT_INTERACTION_ID
        An external unique id the user can set, this id can be used later on to annotate the interaction
        if EXT_INTERACTION_ID was not supplied by the user, deepchecks will try to capture openai response id
        (i.e. - {"id": <openai unique id>, ...} and will set it as the "ext_interaction_id" of the logged interaction
    """
    USER_INPUT = "user_input"
    INFORMATION_RETRIEVAL = "information_retrieval"
    USER_ID = "user_id"
    EXT_INTERACTION_ID = "ext_interaction_id"


class EnvType(str, enum.Enum):
    PROD = "PROD"
    EVAL = "EVAL"


class AnnotationType(str, enum.Enum):
    GOOD = "good"
    BAD = "bad"
    UNKNOWN = "unknown"


@dataclass
class GoldenSetInteraction:
    id: uuid.UUID
    ext_interaction_id: str
    user_input: str
    information_retrieval: str
    full_prompt: str
    response: str
    created_at: datetime


class StepType(str, enum.Enum):
    LLM = 'LLM'
    INFORMATION_RETRIVAL = 'INFORMATION_RETRIVAL'
    TRANSFORMATION = 'TRANSFORMATION'
    FILTER = 'FILTER'
    FINE_TUNING = 'FINE_TUNING'
    PII_REMOVAL = 'PII_REMOVAL'
    UDF = 'UDF'


@dataclass
class Step:
    name: str
    type: StepType
    attributes: Dict[str, Any]
    started_at: datetime = datetime.now()
    annotation: Union[AnnotationType, None] = None
    finished_at: Union[datetime, None] = None
    input: Union[str, None] = None
    output: Union[str, None] = None
    error: Union[str, None] = None

    def to_json(self):
        return {
            'name': self.name,
            'annotation': self.annotation.value if self.annotation is not None else None,
            'type': self.type.value,
            'attributes': self.attributes,
            'started_at': self.started_at.astimezone().isoformat() if self.started_at else None,
            'finished_at': self.finished_at.astimezone().isoformat() if self.finished_at else None,
            'input': self.input,
            'output': self.output,
            'error': self.error
        }

    @classmethod
    def as_jsonl(cls, steps):
        if steps is None:
            return None
        return [step.to_json() for step in steps]



import dunamai as _dunamai
from . import (
    api,
    client,
    data_types,
    openai_instrumentor,
    utils
)

__version__ = _dunamai.get_version("deepchecks_llm_client",
                                   third_choice=_dunamai.Version.from_any_vcs).serialize(style=_dunamai.Style.Pep440)





import functools
import logging
import json
import typing as t
from json import JSONDecodeError

import httpx


def maybe_raise(
        response: httpx.Response,
        expected: t.Union[int, t.Tuple[int, int]] = (200, 299),
        msg: t.Optional[str] = None
) -> httpx.Response:
    """Verify response status and raise an HTTPError if got unexpected status code.

    Parameters
    ==========
    response : Response
        http response instance
    expected : Union[int, Tuple[int, int]] , default (200, 299)
        HTTP status code that is expected to receive
    msg : Optional[str] , default None
        error message to show in case of unexpected status code,
        next template parameters available:
        - status (HTTP status code)
        - reason (HTTP reason message)
        - url (request url)
        - body (response payload if available)
        - error (default error message that will include all previous parameters)

    Returns
    =======
    Response
    """
    status = response.status_code
    url = response.url
    reason = response.content

    error_template = 'Error: {status} {reason} url {url}.\nBody:\n{body}'
    client_error_template = '{status} Client Error: {reason} for url: {url}.\nBody:\n{body}'

    server_error_template = (
        '{status} Server Internal Error: {reason} for url: {url}.\n'
        'Please, reach Deepchecks support team for more information about this problem.\n'
        'Body:\n{body}'
    )

    def select_template(status):
        if 400 <= status <= 499:
            return client_error_template
        elif 500 <= status <= 599:
            return server_error_template
        else:
            return error_template

    def process_body():
        try:
            return json.dumps(response.json(), indent=3)
        except JSONDecodeError:
            return

    if isinstance(expected, int) and status != expected:
        body = process_body()
        error = select_template(status).format(status=status, reason=reason, url=url, body=body)
        raise httpx.HTTPStatusError(
            error if msg is None else msg.format(
                status=status,
                reason=reason,
                url=url,
                body=body,
                error=error
            ),
            request=response.request,
            response=response
        )

    if isinstance(expected, (tuple, list)) and not expected[0] <= status <= expected[1]:
        body = process_body()
        error = select_template(status).format(status=status, reason=reason, url=url, body=body)
        raise httpx.HTTPStatusError(
            error if msg is None else msg.format(
                status=status,
                reason=reason,
                url=url,
                body=body,
                error=error
            ),
            request=response.request,
            response=response

        )
    return response


def null_log_filter(record):
    return False


def set_verbosity(verbose, logger: logging.Logger):
    if not verbose:
        logger.addFilter(null_log_filter)


def handle_exceptions(logger: logging.Logger, return_self: bool = False):
    def decorator(original_func):
        @functools.wraps(original_func)
        def wrapper(*args, **kwargs):
            try:
                return original_func(*args, **kwargs)
            except Exception as ex:
                logger.error("Deepchecks SDK encountered a problem in %s: %s", original_func.__name__, str(ex))
                if return_self:
                    return args[0]
                else:
                    return None
        return wrapper
    return decorator





import functools
import logging
import time

import openai
from deepchecks_llm_client.api import API
from deepchecks_llm_client.utils import handle_exceptions, set_verbosity

logging.basicConfig()
logger = logging.getLogger(__name__)


class OpenAIInstrumentor:

    def __init__(self, api: API, verbose: bool = False, log_level: int = logging.WARNING):
        self.api = api

        logger.setLevel(log_level)
        set_verbosity(verbose, logger)

        self.openai_version = "0.0.0"
        try:
            from importlib import metadata
            self.openai_version = metadata.version('openai')
        except Exception as ex:
            pass

    @staticmethod
    def _patched_call(original_fn, patched_fn):
        @functools.wraps(original_fn)
        def _inner_patch(*args, **kwargs):
            return patched_fn(original_fn, *args, **kwargs)
        return _inner_patch

    def patcher_create(self, original_fn, *args, **kwargs):

        self._before_run_log_print(args, kwargs, original_fn)

        timestamp = time.time()
        result = original_fn(*args, **kwargs)
        time_delta = time.time() - timestamp

        self._after_run_actions(args, kwargs, original_fn, result, time_delta)

        return result

    async def patcher_acreate(self, original_fn, *args, **kwargs):

        self._before_run_log_print(args, kwargs, original_fn)

        timestamp = time.time()
        result = await original_fn(*args, **kwargs)
        time_delta = time.time() - timestamp

        self._after_run_actions(args, kwargs, original_fn, result, time_delta)

        return result

    @handle_exceptions(logger)
    def _after_run_actions(self, args, kwargs, original_fn, result, time_delta):
        logger.info("Finished running function: %s, time delta: %s", original_fn.__qualname__, time_delta)
        logger.debug("Function Response: %s", result)

        # Obfuscate the api-key
        if kwargs.get("api_key"):
            kwargs["api_key"] = f"last-4-digits-{kwargs['api_key'][-4:]}"

        event_dict = {
            "request": {"func_name": original_fn.__qualname__, "args": args, "kwargs": kwargs},
            "response": result.to_dict_recursive(),
            "runtime_data": {"response_time": time_delta, "openai_version": self.openai_version},
        }
        self.api.load_openai_data(data=[event_dict])

        logger.debug("Reported this event to deepchecks server:\n%s", event_dict)

    @staticmethod
    def _before_run_log_print(args, kwargs, original_fn):
        logger.info("Running the original function: %s. args: %s, kwargs: %s", original_fn.__qualname__, args, kwargs)

    def perform_patch(self):
        try:
            openai.ChatCompletion.acreate = self._patched_call(
                openai.ChatCompletion.acreate, self.patcher_acreate
            )
        except AttributeError:
            pass

        try:
            openai.ChatCompletion.create = self._patched_call(
                openai.ChatCompletion.create, self.patcher_create
            )
        except AttributeError:
            pass

        try:
            openai.Completion.acreate = self._patched_call(
                openai.Completion.acreate, self.patcher_acreate
            )
        except AttributeError:
            pass

        try:
            openai.Completion.create = self._patched_call(
                openai.Completion.create, self.patcher_create
            )
        except AttributeError:
            pass




import logging
import typing as t
import uuid

from deepchecks_llm_client.api import API
from deepchecks_llm_client.data_types import Tag, EnvType, AnnotationType, GoldenSetInteraction, Step
from deepchecks_llm_client.openai_instrumentor import OpenAIInstrumentor
from deepchecks_llm_client.utils import handle_exceptions, set_verbosity

__all__ = ["dc_client", "DeepchecksLLMClient"]

logging.basicConfig()
logger = logging.getLogger(__name__)
init_logger = logging.Logger(__name__ + ".init")

DEFAULT_VERSION_NAME = '0.0.1'
DEFAULT_ENV_TYPE = EnvType.PROD


class DeepchecksLLMClientError(Exception):
    pass


class DeepchecksLLMClient:
    def __init__(self):
        self._api: API = None
        self.app: t.Dict[str, t.Any] = None
        self.instrumentor: OpenAIInstrumentor = None

    @handle_exceptions(init_logger)
    def init(self,
             host: str,
             api_token: str,
             app_name: str,
             version_name: str = DEFAULT_VERSION_NAME,
             env_type: EnvType = DEFAULT_ENV_TYPE,
             auto_collect: bool = True,
             init_verbose: bool = True,
             verbose: bool = False,
             log_level: int = logging.WARNING
             ):
        """
        Connect to Deepchecks LLM Server

        Parameters
        ==========
        host : str
            Deepchecks host to communicate with
        api_token : str
            Deepchecks API Token (can be generated from the UI)
        app_name : str
            Application name to connect to
        version_name : str, default='0.0.1'
            Version name to connect to inside the application,
            if Version name does not exist SDK will create it automatically,
        env_type : EnvType, default=EnvType.PROD
            could be EnvType.PROD (for 'Production') or EnvType.EVAL (for 'Evaluation')
        auto_collect : bool, default=True
            Auto collect calls to LLM Models
        init_verbose: bool, default=True
            Write log messages during the init phase and non-prod phases such as golden_set()
        verbose : bool, default=False
            Write log messages (by default we are in silence mode)
        log_level: int, default=logging.WARNING
            In case that verbose or init_verbose is True,
            this parameter will set SDK loggers logging level

        Returns
        =======
        None
        """
        logger.setLevel(log_level)
        set_verbosity(init_verbose, init_logger)
        set_verbosity(verbose, logger)

        if self._api is None:
            if host is not None and api_token is not None:
                self._api = API.instantiate(host=host, token=api_token)
            else:
                raise DeepchecksLLMClientError('host/token parameters must be provided')

        if app_name is None:
            raise DeepchecksLLMClientError('app_name must be supplied')
        self.app = self._api.get_application(app_name)
        if not self.app:
            raise DeepchecksLLMClientError(f'Application: "{app_name}", does not exist, please create it via the UI')

        self.app_name(app_name).version_name(version_name).env_type(env_type)

        self.instrumentor = None
        if auto_collect:
            self.instrumentor = OpenAIInstrumentor(self.api, verbose, log_level)
            self.instrumentor.perform_patch()

    @property
    @handle_exceptions(logger)
    def api(self) -> API:
        if self._api:
            return self._api
        raise DeepchecksLLMClientError("dc_client was not initialized correctly, please re-create it")

    @handle_exceptions(logger, return_self=True)
    def app_name(self, new_app_name: str):
        self.api.app_name(new_app_name)
        return self

    @handle_exceptions(logger, return_self=True)
    def version_name(self, new_version_name: str):
        self.api.version_name(new_version_name)
        return self

    @handle_exceptions(logger, return_self=True)
    def env_type(self, new_env_type: EnvType):
        self.api.env_type(new_env_type)
        return self

    @handle_exceptions(logger)
    def set_tags(self, tags: t.Dict[Tag, str]):
        self.api.set_tags(tags)
        return self

    @handle_exceptions(logger)
    def annotate(self, ext_interaction_id: str, annotation: AnnotationType):
        self.api.annotate(ext_interaction_id, annotation)
        return self

    @handle_exceptions(logger)
    def log_interaction(self,
                        user_input: str,
                        model_response: str,
                        full_prompt: str = None,
                        information_retrieval: str = None,
                        annotation: AnnotationType = None,
                        ext_interaction_id: str = None,
                        steps: t.List[Step] = None,
                        custom_props: t.Dict[str, t.Any] = None) -> t.Union[str, uuid.UUID, None]:
        result = self.api.log_interaction(user_input=user_input, model_response=model_response,
                                          full_prompt=full_prompt, information_retrieval=information_retrieval,
                                          annotation=annotation, ext_interaction_id=ext_interaction_id, steps=steps,
                                          custom_props=custom_props)
        return result.json()

    @handle_exceptions(logger)
    def update_application_config(self, file):
        self.api.update_application_config(application_id=self.app['id'], file=file)

    @handle_exceptions(logger)
    def get_application_config(self, file_save_path: t.Union[str, None] = None) -> str:
        return self.api.get_application_config(application_id=self.app['id'], file_save_path=file_save_path)

    @handle_exceptions(init_logger)
    def golden_set(self, version_name: t.Union[str, None] = None) -> t.Iterable[GoldenSetInteraction]:
        """
        Fetch all that interactions from the golden set (EnvType.EVAL)

        Parameters
        ----------
        version_name : str
            version name to fetch interactions from, if no version name was supplied
            the latest created version in the application will be used (mostly useful
            when integrating the golden set rerun in the CI)

        Returns
        -------
        Iterable collection of interactions

        """
        golden_set_version = self._get_golden_set_version(version_name)

        offset = 0
        limit = 20

        while True:
            interactions = self.api.get_interactions(golden_set_version["id"], limit=limit, offset=offset)
            for interaction in interactions:
                yield GoldenSetInteraction(id=interaction["id"],
                                           ext_interaction_id=interaction.get("ext_interaction_id"),
                                           user_input=interaction["input"]["data"] if interaction.get("input") else None,
                                           information_retrieval=interaction["information_retrieval"]["data"] if interaction.get("information_retrieval") else None,
                                           full_prompt=interaction["prompt"]["data"] if interaction.get("prompt") else None,
                                           response=interaction["response"]["data"] if interaction.get("response") else None,
                                           created_at=interaction["created_at"])

            # If the size of the data is less than the limit, we've reached the end
            if len(interactions) < limit:
                break

            offset += limit

    @handle_exceptions(init_logger)
    def golden_set_generation_completed(self):
        """
        This function indicate deepchecks that golden set generation completed
        It helps to finalize golden set related actions, for instance: train topics
        model and extract topics (the model will be used later on to extract topics
        from production data)

        Returns
        -------
        None

        """
        self.api.generate_topics()

    def _get_golden_set_version(self, version_name):

        app = self.api.get_application(self.api.get_app_name())
        if not app:
            raise DeepchecksLLMClientError(f"Application: '{self.api.get_app_name()}', does not exist")
        if version_name:
            golden_set_version = next((ver for ver in app["versions"] if ver["name"] == version_name), None)
            if not golden_set_version:
                raise DeepchecksLLMClientError(f"Could not find version '{version_name}', in application '{self.api.get_app_name()}'")
        else:
            golden_set_version = max(app["versions"], key=lambda x: x['created_at'])
            if not golden_set_version:
                raise DeepchecksLLMClientError(f"Could not find versions to select from in application '{self.api.get_app_name()}'")
        return golden_set_version


# LLM Client publicly accessed singleton
dc_client: DeepchecksLLMClient = DeepchecksLLMClient()



# -----------------------------------------------------------------------------
# Deepchecks LLM Onboarding Script
# -----------------------------------------------------------------------------
# This script is designed to guide you through the process of integrating
# data with Deepchecks LLM Product, using Deepchecks' SDK.
# Please follow the instructions provided and replace placeholders with actual
# data as needed.


# Step 1: Import necessary modules
# Create new venv and install deepchecks' SDK using `pip install deepchecks-llm-client`
import uuid
from datetime import datetime

import openai

from deepchecks_llm_client.client import dc_client
from deepchecks_llm_client.data_types import Tag, EnvType, AnnotationType, Step, StepType

# Step 2: Set up configuration

# Fill deepchecks host name here
DEEPCHECKS_LLM_HOST = "https://host-name-here"  # could be: https://app.llm.deepchecks.com

# Login to deepchecks' service and generate new API Key (Configuration -> API Key) and place it here
DEEPCHECKS_LLM_API_KEY = "Fill Key Here"

# Use "Update Samples" in deepchecks' service, to create a new application name and place it here
DEEPCHECKS_APP_NAME = "DemoApp"

# In case you would like to experience with deepchecks auto collect capabilities, please set
# Open AI's key here (as string)
OPENAI_API_KEY = "Fill Key Here"

# Step 3: Initialize the SDK

# auto_collect=True wraps `openai.ChatCompletion` and `openai.Completion` APIs
# so any OpenAI invocation will fire an event to Deepchecks with the relevant data
# If you do not plan to use OpenAI auto collection, you can set it to `False`

# Notice - deepchecks SDK was designed to be non-intrusive. It does not throw exceptions
# and only print to log. By default, it prints only errors from the "init" phase
# If you wish to increase verbosity, use `verbose=True` and `log_level=logging.INFO`
version = "0.0.1"
env_type = EnvType.EVAL
print(f"init deepchecks client, host: {DEEPCHECKS_LLM_HOST}, app: {DEEPCHECKS_APP_NAME}, version: {version}, env: {env_type}")
dc_client.init(host=DEEPCHECKS_LLM_HOST,
               api_token=DEEPCHECKS_LLM_API_KEY,
               app_name=DEEPCHECKS_APP_NAME,
               version_name=version,
               env_type=env_type,
               auto_collect=True)

# Step 4: Auto collecting OpenAI Data (Q&A + Annotation)
if OPENAI_API_KEY:

    # You can set_tags() to add additional information to the collected data
    # most important is the USER_INPUT, which is the actual user question, without this
    # information, deepchecks won't be able to evaluate the question the user asked
    USER_INPUT = "How much is 2 plus 2?"
    # If you have "information retrieval" step in your flow, you can supply is as well
    # using the Tag.INFORMATION_RETRIEVAL tag
    INFORMATION_RETRIEVAL = "1 plus 1 equals 2"
    # Notice that you can add USER_ID (optional) to track the number of unique users that are
    # using your LLM application (remember PII and obfuscate the username if needed)
    dc_client.set_tags({Tag.USER_INPUT: USER_INPUT,
                        Tag.USER_ID: "A05fdfbb2035e@gmail.com",
                        Tag.INFORMATION_RETRIEVAL: INFORMATION_RETRIEVAL})

    # Setting up your OpenAI API credentials
    print("calling OpenAI ChatCompletion API")
    openai.api_key = OPENAI_API_KEY
    chat_completion = openai.ChatCompletion.create(model="gpt-3.5-turbo",
                                                   temperature=0.7,
                                                   messages=[
                                                       {"role": "system", "content": f"Act as a calculator {INFORMATION_RETRIEVAL}"},
                                                       {"role": "user", "content": USER_INPUT}])

    # The data has already logged to deepchecks server, you can annotate the data
    # in anytime you like using OpenAI's unique id
    # Notice that you can set your own unique id using Tags.EXT_INTERACTION_ID (before calling
    # OpenAI's API), and use it here instead of OpenAI's id
    print(f"Annotating OpenAI interaction with key: {chat_completion.openai_id}")
    dc_client.annotate(chat_completion.openai_id, annotation=AnnotationType.GOOD)

    # clear tags so we won't carry them on to the next calls
    # relevant only for "auto collected" data
    dc_client.set_tags({})
else:
    print("OpenAI Key was not supply, skipping OpenAI demo")

# Step 5: Explicit Logging of data and optionally addition of steps for flow visibility
# Notice that we can also log custom properties - you can use custom properties that were
# already defined via the UI, or use the "CP_xxx" convention to create new custom property
# on the fly

# We can explicitly log interaction
print(f"log interaction using explicit Deepchecks API: log_interaction()")
ext_id = str(uuid.uuid4())
dc_client.log_interaction(user_input="my user input",
                          model_response="my model response",
                          full_prompt="system part: my user input",
                          information_retrieval="system part: my information retrieval",
                          annotation=AnnotationType.BAD,
                          ext_interaction_id=ext_id,
                          steps=[
                              Step(
                                  name="Information retrival",
                                  type=StepType.INFORMATION_RETRIVAL,
                                  attributes={"env": "dev"}),
                              Step(
                                  input="my user input + the retried data",
                                  name="Information retrival",
                                  type=StepType.LLM,
                                  attributes={"env": "dev"},
                                  annotation=AnnotationType.BAD,
                                  output="my model response"),
                              Step(
                                  name="Information retrival",
                                  type=StepType.TRANSFORMATION,
                                  attributes={"env": "dev"},
                                  output="my model response after PII removal")
                          ],
                          custom_props={"CP_MY_PROP1": 0.1})

# Notice that we can re-annotate our logged interation in a later phase
# using out own external id
print(f"Annotating explicit interaction with key: {ext_id}")
dc_client.annotate(ext_id, AnnotationType.GOOD)

# Indicate we finished process golden set data for 0.0.1
dc_client.golden_set_generation_completed()

# Step 6: Creating a new version based on the "0.0.1" golden set

# Set different version / env type, to be carried on to the next calls
# Usually, in "dev mode", you will log to "EnvType.EVAL" and use a different version
# for each evaluation you make. Check the performance of your LLM application using deepchecks,
# and deploy the selected version to production, and from that point on,
# you will use "EnvType.PROD" with the "production version" to log production data
new_version = "0.0.2"
eval_env_type = EnvType.EVAL
print(f"switching to version: {new_version}, and EnvType: {eval_env_type}")
dc_client.version_name(new_version).env_type(eval_env_type)

# Now we have 1 or 2 interactions logged into deepchecks' under version '0.0.1'
# We can get the "golden set" out of deepchecks and rerun it against OpenAI while log the
# new result to our new version (i.e. - "0.0.2")
if OPENAI_API_KEY:

    # golden_set() can get "version_name" as param, if no version_name was supplied
    # the latest created version will be taken (this is useful when integrating it in the CI)
    for interaction in dc_client.golden_set(version_name="0.0.1"):
        print(f"processing golden set interaction with ext_interaction_id: {interaction.ext_interaction_id}  "
              f"(deepchecks id: {interaction.id}, created_at: {interaction.created_at})")
        dc_client.set_tags({Tag.USER_INPUT: interaction.user_input,
                            Tag.INFORMATION_RETRIEVAL: interaction.information_retrieval})

        # We run the same OpenAI command, only changing the temperature...
        chat_completion = openai.ChatCompletion.create(model="gpt-3.5-turbo",
                                                       api_key=OPENAI_API_KEY,
                                                       temperature=0.2,
                                                       messages=[
                                                           {"role": "system", "content": f"Act as a calculator {interaction.information_retrieval}"},
                                                           {"role": "user", "content": interaction.user_input}])

    # Indicate we finished process golden set data for 0.0.2
    dc_client.golden_set_generation_completed()



import logging
import os
from collections import defaultdict
from datetime import datetime
from typing import Any, Union, Dict, List, Optional, Sequence
from uuid import UUID

from langchain.callbacks.base import BaseCallbackHandler
from langchain.schema import LLMResult
from langchain.schema.document import Document
from deepchecks_llm_client.client import DeepchecksLLMClient
from deepchecks_llm_client.data_types import Step, StepType, EnvType

from deepchecks_llm_client.client import DEFAULT_ENV_TYPE


class DeepchecksCallbackHandler(BaseCallbackHandler):
    """Base callback handler that can be used to handle callbacks from langchain."""

    def __init__(self, *args, app_name, app_version, env_type: EnvType = DEFAULT_ENV_TYPE,
                 host: str = None, api_key: str = None, logger: logging.Logger = None,
                 ext_interaction_id: Union[str, None] = None,
                 **kwargs):
        super().__init__(*args, **kwargs)
        self.api_key = api_key
        self.host = host
        self.app_name = app_name
        self.app_version = app_version
        self.env_type = env_type
        self.logger = logger or logging.getLogger(__name__)
        self._ext_interaction_id = ext_interaction_id
        self._user_input = None
        self._user_response = None
        self._steps: Dict[str, Step] = defaultdict()  # run_id -> Step

    @property
    def steps(self) -> List[Step]:
        return list(self._steps.values())

    @property
    def user_input(self) -> str:
        return self._user_input

    @property
    def user_response(self) -> str:
        return self._user_response

    @property
    def ext_interaction_id(self) -> Union[str, None]:
        return str(self._ext_interaction_id)

    def on_llm_start(
            self, serialized: Dict[str, Any], prompts: List[str], **kwargs: Any
    ) -> Any:
        """Run when LLM starts running."""
        self.logger.debug('on_llm_start', kwargs)
        run_id = kwargs['run_id']

        self._steps[str(run_id)] = Step(
            name="LLM",
            type=StepType.LLM,
            attributes=self.to_str_dict(kwargs),
            started_at=datetime.now().astimezone(),
            input='\n'.join(prompts),
        )

    def on_llm_end(self, response: LLMResult, **kwargs: Any) -> Any:
        """Run when LLM ends running."""
        self.logger.debug('on_llm_end', kwargs)
        step = self._steps[str(kwargs['run_id'])]
        step.finished_at = datetime.now().astimezone()
        step.output = '\n'.join([a.json() for b in response.generations for a in b])

    def on_llm_error(
            self, error: Union[Exception, KeyboardInterrupt], **kwargs: Any
    ) -> Any:
        """Run when LLM errors."""
        self.logger.debug('on_llm_error', kwargs)
        self._on_error(error, kwargs)

    def on_chain_start(
            self,
            serialized: Dict[str, Any],
            inputs: Dict[str, Any],
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            tags: Optional[List[str]] = None,
            metadata: Optional[Dict[str, Any]] = None,
            **kwargs: Any,
    ) -> None:
        """Run when chain starts running."""
        if parent_run_id is None:
            self.logger.debug('on_chain_start', kwargs)
            self._user_input = inputs.get('input') or inputs.get('question') or inputs.get('text') or '\n'.join([str(v) for v in inputs.values()]) \
                if isinstance(inputs, dict) \
                else inputs

    def on_chain_end(self, outputs: Dict[str, Any], **kwargs: Any) -> Any:
        """Run when chain ends running."""
        self.logger.debug('on_chain_end', kwargs)
        if kwargs.get('parent_run_id', None) is None:
            self._user_response = outputs.get('text') or outputs.get('output_text') or '\n'.join([str(v) for v in outputs.values()]) \
                if isinstance(outputs, dict) \
                else outputs

            self._log_interaction()

    def on_retriever_error(
            self,
            error: BaseException,
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any,
    ) -> Any:
        """Run when Retriever errors."""
        self.logger.debug('on_retriever_error', kwargs)
        self._on_error(error, kwargs)

    def _on_error(self, error, kwargs):
        if step := self._steps.get(str(kwargs['run_id'])):
            step.finished_at = datetime.now().astimezone()
            step.error = str(error)
        self._log_interaction()

    def on_retriever_start(
            self,
            serialized: Dict[str, Any],
            query: str,
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            tags: Optional[List[str]] = None,
            metadata: Optional[Dict[str, Any]] = None,
            **kwargs: Any,
    ) -> Any:
        """Run when Retriever starts running."""
        self.logger.debug('on_retriever_start', kwargs)
        self._steps[str(run_id)] = Step(
            name="Data Retriever",
            input=str(query),
            type=StepType.INFORMATION_RETRIVAL,
            attributes=self.to_str_dict(kwargs),
            started_at=datetime.now().astimezone(),

        )

    def on_retriever_end(
            self,
            documents: Sequence[Document],
            *,
            run_id: UUID,
            parent_run_id: Optional[UUID] = None,
            **kwargs: Any,
    ) -> Any:
        """Run when Retriever ends running."""
        self.logger.debug('on_retriever_end', kwargs)
        self._steps[str(run_id)].finished_at = datetime.now().astimezone()
        self._steps[str(run_id)].output = '\n'.join(d.page_content for d in documents)

    def _log_interaction(self):
        _client: DeepchecksLLMClient = DeepchecksLLMClient()
        _client.init(host=os.getenv('DEEPCHECKS_LLM_HOST', self.host),
                     api_token=os.getenv('DEEPCHECKS_LLM_API_KEY', self.api_key),
                     app_name=self.app_name,
                     version_name=self.app_version,
                     env_type=self.env_type,
                     auto_collect=False)
        result = _client.log_interaction(
            model_response=self.user_response,
            ext_interaction_id=self.ext_interaction_id,
            user_input=self.user_input,
            steps=self.steps,
            information_retrieval=None,
        )
        if not result:
            raise self.logger.error('Interaction logging failed')

    @staticmethod
    def to_str_dict(attributes):
        return {k: str(v) for k, v in attributes.items()}



import logging
import typing as t
import uuid
import warnings
from copy import copy

import httpx
import packaging.version
from httpx import URL

import deepchecks_llm_client
from deepchecks_llm_client.data_types import Tag, EnvType, AnnotationType, Step
from deepchecks_llm_client.utils import maybe_raise

__all__ = ['API']


logger = logging.getLogger(__name__)

TAPI = t.TypeVar('TAPI', bound='API')


class API:

    session: httpx.Client
    original_host: URL

    @classmethod
    def instantiate(cls: t.Type[TAPI], host: str, token: t.Optional[str] = None) -> TAPI:
        headers = {'Authorization': f'Basic {token}'} if token else None
        return cls(session=httpx.Client(base_url=host, headers=headers, timeout=60))

    def __init__(self, session: httpx.Client):
        self.session = copy(session)
        self.original_host = self.session.base_url
        self.session.base_url = self.session.base_url.join('/api/v1')
        self._app_name: str = None
        self._version_name: str = None
        self._env_type: EnvType = None
        self._tags: t.Dict[Tag, str] = {}

        try:
            backend_version = packaging.version.parse(self.retrieve_backend_version())
            client_version = packaging.version.parse(deepchecks_llm_client.__version__)
        except packaging.version.InvalidVersion as ex:
            raise Exception(f'Not able to compare backend and client versions, '
                            f'backend or client use incorrect or legacy versioning schema: {str(ex)}')
        else:
            if backend_version.major != client_version.major:
                raise Exception(
                    f'You are using an old, potentially incompatible with the current API, client version.\n'
                    f'Client version is {client_version}, Backend version is {backend_version}\n'
                    f'Upgrade "deepchecks-llm-client" version by running:\n'
                    f'>> pip install -U deepchecks-llm-client')

    def app_name(self, new_app_name: str):
        if new_app_name is None:
            raise ValueError("new_app_name cannot be set to None")
        self._app_name = new_app_name

    def get_app_name(self):
        return self._app_name

    def version_name(self, new_version_name: str):
        if new_version_name is None:
            raise ValueError("new_version_name cannot be set to None")
        self._version_name = new_version_name

    def get_version_name(self):
        return self._version_name

    def env_type(self, new_env_type: EnvType):
        if new_env_type is None:
            raise ValueError("new_env_type cannot be set to None")
        if new_env_type not in (EnvType.PROD, EnvType.EVAL):
            raise ValueError("new_env_type must be one of: EnvType.PROD or EnvType.EVAL")
        self._env_type = new_env_type

    def get_env_type(self):
        return self._env_type

    def set_tags(self, tags: t.Dict[Tag, str]):
        if tags is None:
            self._tags = {}
        else:
            self._tags = tags

    def retrieve_backend_version(self) -> str:
        payload = maybe_raise(self.session.get('backend-version')).json()
        return payload['version']

    def get_application(self, app_name: t.Union[str, None] = None) -> t.Dict[str, t.Any]:
        if app_name is None:
            app_name = self._app_name
        payload = maybe_raise(self.session.get('applications', params={"name": [app_name]})).json()

        return payload[0] if len(payload) > 0 else None

    def load_openai_data(self,  data: t.List[t.Dict[str, t.Any]]) -> t.Optional[httpx.Response]:
        for row in data:
            row["user_data"] = self._tags
        return maybe_raise(self.session.post('openai-load', json=data,
                                             params={'app_name': self._app_name,
                                                     'version_name': self._version_name,
                                                     'env_type': self._env_type.value}))

    def annotate(self, ext_interaction_id: str, annotation: AnnotationType) -> t.Optional[httpx.Response]:
        return maybe_raise(self.session.post('annotations', json={"ext_interaction_id": ext_interaction_id,
                                                                  "value": annotation.value}))

    def log_interaction(self, user_input: str, model_response: str, full_prompt: str,
                        information_retrieval: str, annotation: AnnotationType,
                        ext_interaction_id: str, steps: t.List[Step],
                        custom_props: t.Dict[str, t.Any]) -> t.Optional[httpx.Response]:
        return maybe_raise(self.session.post('interactions',
                                             json={"ext_interaction_id": ext_interaction_id,
                                                   "user_input": user_input,
                                                   "model_response": model_response,
                                                   "full_prompt": full_prompt,
                                                   "information_retrieval": information_retrieval,
                                                   "app_name": self._app_name,
                                                   "version_name": self._version_name,
                                                   "env_type": self._env_type.value,
                                                   "annotation": annotation.value if annotation else None,
                                                   "raw_json_data": {"user_data": self._tags},
                                                   "steps": Step.as_jsonl(steps),
                                                   "custom_props": custom_props}), expected=201)

    def get_interactions(self, application_version_id: int, limit: int, offset: int) -> t.List:
        return maybe_raise(
            self.session.post('get-interactions-by-filter',
                              json={
                                  "application_version_id": application_version_id,
                                  "environment": EnvType.EVAL.value,
                                  "limit": limit,
                                  "offset": offset
                              })
        ).json()

    def generate_topics(self):
        maybe_raise(self.session.post(f'generate-topics?env_type={self.get_env_type().value}'
                                      f'&app_name={self.get_app_name()}&app_version_name={self.get_version_name()}'))

    def update_application_config(self, application_id: int, file):
        if isinstance(file, str):
            file = open(file, "rb")
        data = {"file": ("filename", file)}
        maybe_raise(self.session.put(f'applications/{application_id}/config', files=data))

    def get_application_config(self, application_id: int, file_save_path: t.Union[str, None] = None) -> str:
        text = maybe_raise(self.session.get(f'applications/{application_id}/config')).text
        if file_save_path:
            with open(file_save_path, "w") as f:
                f.write(text)
        return text



import pathlib
import typing as t

import setuptools
from dunamai import Version

SETUP_MODULE = pathlib.Path(__file__).absolute()
ROOTDIR = SETUP_MODULE.parent


def read_requirements_file(path: pathlib.Path) -> t.Tuple[t.List[str], t.List[str]]:
    if not (path.exists() and path.is_file()):
        raise RuntimeError(f'Did not find requirements file - {path.name}')
    dependencies = []
    dependencies_links = []
    for line in path.open("r").readlines():
        if "-f" in line or "--find-links" in line:
            dependencies_links.append(
                line
                .replace("-f", "")
                .replace("--find-links", "")
                .strip()
            )
        else:
            dependencies.append(line)
    return dependencies, dependencies_links


# ===============================================================

install_requires, dependency_links = read_requirements_file(ROOTDIR / 'requirements.txt')


setuptools.setup(
    name="deepchecks-llm-client",
    version="0.5.6",
    author="deepchecks",
    author_email="info@deepchecks.com",
    description="The SDK client for communicating with Deepchecks LLM service",
    packages=setuptools.find_packages(where=".", include=["deepchecks_llm_client", "deepchecks_llm_client.*"]),
    python_requires='>=3.8',
    install_requires=install_requires,
    dependency_links=dependency_links,
    include_package_data=True,
)